module BidiEngine where

import IntermediateTransfType
import Data.List
import Types

-- How rule inversion works:
-- 1) Extract the canBes from the tranformation
-- 2) Do one fast rule inversion, and remove all the rules that were used
--    to generate the canBes. We don't need those to generate the backwards
--    transformation
-- 3) Use the canBes and extend all the rules as much as possible to target
--    all possible non terminals
-- 4) Take the result of 3) and create an IntermediateTransf, from which we
--    generate Silver code
-- 5) Go through the rules rhs and change the P_Arg to P_Put. This is not simple,
--    and requires going to the lhs of the rule to see the correct types I want
--    to map back to
-- *) There could be a last step, convert all the P_Get to P_Put on the lhs of
--    a rule. We don't do this here because it is done when generating the
--    Silver code.

invert :: (Ord prodNameA, Ord nonTermA, Ord termA, Ord prodNameB, Ord nonTermB, Ord termB )
       => Transf termA nonTermA prodNameA termB nonTermB prodNameB
       -> Transf termB nonTermB prodNameB termA nonTermA prodNameA
invert t@(Transf rules) = let -- First thing, compute the list of canBes
                              canBes    = extractCanBes t
                              -- Do one first quick inversion
                              inverted  = map (\(a,b) -> (b,a)) rules
                              -- Remove the rules which generate canBes. We don't need these.
                              cleaned1   = filter (not . isPatternCanBe . fst) $ inverted
                              -- And we'll remove the P_Get on all the rhs rules.
                              cleaned2   = map (\(a,(P_Get _ p)) -> (a,p)) cleaned1
                              -- And transform the lhs of rules into P_Put
                              toPut      = map (\(a,b) -> (P_Put (getPatternType b) a, b)) cleaned2
                              -- Expansion using canBes
                              expansion = concat $ map (expandRules canBes) toPut
                              -- Create P_Put on the rhs
                              withP_Put = map makeP_Put expansion
                          in  Transf withP_Put
                          where makeP_Put (a, b) = (a, search_right a b)

-- Do step 5) is harder than it looks, because we have to jump between
-- patterns of differente types to search everything (that is what the)
-- aux functions used for, and then we have to check if the variable is in
-- the scope of the P_Get or P_Put to see if their type is the one to be used
search_right :: Pattern termB nonTermB prodNameB termA nonTermA prodNameA
             -> Pattern termB nonTermB prodNameB termA nonTermA prodNameA
             -> Pattern termB nonTermB prodNameB termA nonTermA prodNameA
search_right lhs (P_Get t p)       = P_Get t (aux_right lhs p)
search_right lhs (P_Put t p)       = P_Put t (aux_right lhs p)
search_right lhs (Pattern p n rhs) = Pattern p n (map (search_right lhs) rhs)
search_right lhs (P_Arg n t)       = let aux = (\(P_Put _ p) -> p) lhs
                                     in  P_Put t (P_Arg n (head $ search_left n [aux]))
search_right _   p                 = p
aux_right :: Pattern termB nonTermB prodNameB termA nonTermA prodNameA
             -> Pattern termA nonTermA prodNameA termB nonTermB prodNameB
             -> Pattern termA nonTermA prodNameA termB nonTermB prodNameB
aux_right lhs (Pattern p n rhs) = Pattern p n (map (aux_right lhs) rhs) 
aux_right lhs (P_Get t p)       = P_Get t (search_right lhs p)
aux_right lhs (P_Put t p)       = P_Put t (search_right lhs p)

search_left :: Char -> [Pattern termA nonTermA prodNameA termB nonTermB prodNameB] -> [nonTermA]
search_left var (Pattern p n rhs:xs) = (search_left var rhs) ++ (search_left var xs)
search_left var (P_Put t p      :xs) = if (isInScope var p) then [t] else aux_left var p ++ search_left var xs
search_left var (P_Get t p      :xs) = if (isInScope var p) then [t] else aux_left var p ++ search_left var xs
search_left var ( _             :xs) = search_left var xs
search_left _   []                   = []
aux_left var (Pattern p n rhs) = concat $ map (aux_left var) rhs
aux_left var (P_Get t p)       = search_left var [p]
aux_left var (P_Put t p)       = search_left var [p]
aux_left var _                 = []
isInScope var (Pattern p n rhs) = or . map(isInScope var) $ rhs
isInScope var (P_Arg   v n    ) = var==v
isInScope var (P_Get t p      ) = False
isInScope var (P_Put t p      ) = False
isInScope var _                 = False

-- -- -- Rules expansion using the CanBes
expandRules :: Eq nonTermA
            => CanBes termA nonTermA prodNameA termB nonTermB prodNameB
            -> Rule termA nonTermA prodNameA termB nonTermB prodNameB
            -> [Rule termA nonTermA prodNameA termB nonTermB prodNameB]
expandRules canBes@(CanBes l) rule
  = let -- First get the to type from the rhs of a rule
        to             = getPatternType . snd $ rule
        -- Filter the canBes that take those types
        filteredCanBes = filter (\(f,t,_) -> f==to) l
        -- Create the new rules
        newRules       = map (newRule rule) filteredCanBes
    in  rule : newRules
    where newRule rule@(p1,p2) canBe@(_,_,(Pattern prod nt rhs)) = (p1, Pattern prod nt (aux rhs p2))
          aux []                       _  = []
          aux (P_Arg _ _:xs)           p1 = p1:xs
          aux (Pattern prod nt rhs:xs) p1 = (Pattern prod nt (aux rhs p1)):aux xs p1
          aux (x:xs)                   p1 = x:aux xs p1

-- -- -- CanBes
data CanBes termA nonTermA prodNameA termB nonTermB prodNameB = CanBes [(nonTermA, nonTermA, Pattern termA nonTermA prodNameA termB nonTermB prodNameB)]

-- This function will extract canBes from a Transformation using the following steps:
-- 1) Search in the Transformation for rules that only perform jumps between
--    non terminals (i.e., have only one symbol on their rhs which is an
--    argument). These imediatly become canBes.
-- 2) Extend the canbes. For example, if there is a canBe that goes from A -> B
--    and a canBe B -> C we will create a canBe A -> C.
extractCanBes :: (Ord prodNameA, Ord nonTermA, Ord termA, Ord prodNameB, Ord nonTermB, Ord termB)
              => Transf termA nonTermA prodNameA termB nonTermB prodNameB
              -> CanBes termB nonTermB prodNameB termA nonTermA prodNameA
extractCanBes = expandCanBes . searchCanBes

searchCanBes :: Transf termA nonTermA prodNameA termB nonTermB prodNameB -> CanBes termB nonTermB prodNameB termA nonTermA prodNameA
searchCanBes (Transf rules) = let -- Filter the rules whose rhs is a get with only one argument
                                  only1Symbol = filter (isPatternCanBe . snd) rules
                                  -- Now that I have filtered the rules, I really only care about the
                                  -- pattern which is the argument of get on the lhs on the rule
                                  patterns    = map (\(P_Get _ p) -> p) (map fst only1Symbol)
                                  -- Transform these produtions into canBes. This is just to make handle them later easier
                              in  CanBes (map toCanBe patterns)
                              where toCanBe p@(Pattern _ to rhs) = (aux rhs, to, p)
                                    aux (P_Arg _ nt:xs) = nt
                                    aux (_:xs)          = aux xs

-- This function circularly adds new canBes until no more canBes can be computed, i.e., 
-- the old canBes list is equal to the new canBes list
expandCanBes :: (Ord prodNameA, Ord nonTermA, Ord termA, Ord prodNameB, Ord nonTermB, Ord termB)
             => CanBes termA nonTermA prodNameA termB nonTermB prodNameB
             -> CanBes termA nonTermA prodNameA termB nonTermB prodNameB
expandCanBes c@(CanBes l) = let oneTry = concat $ map (expandCanBe c) l
                                -- Because this step probably added repetition
                                -- and we need to sort for comparison
                                new  = nub . sort $ oneTry
                                old  = nub . sort $ l
                            in  if old==new then (CanBes new) else expandCanBes (CanBes new)

expandCanBe :: Eq nonTermA
            => CanBes termA nonTermA prodNameA termB nonTermB prodNameB
            -> (nonTermA, nonTermA, Pattern termA nonTermA prodNameA termB nonTermB prodNameB)
            -> [(nonTermA, nonTermA, Pattern termA nonTermA prodNameA termB nonTermB prodNameB)]
expandCanBe canBes@(CanBes l) canBe@(from, to, patt)
  = let -- First, find CanBes that can be used to expand the current one
        found = filter (\(f, t, p) -> from/=f && to/=t && to==f) l
        -- Secondly, create a list with the new canBes
        new   = map (apply canBe) found
        -- Last, extend the CanBes list ONLY with the ones whose types did not exist
        -- or canBes that go to themselves
    in  addCanBe new l
    where apply (f1, t1, p1) (f2, t2, (Pattern prod nt rhs)) = (f1, t2, Pattern prod nt (aux rhs p1))
          aux (P_Arg _ _:xs)           p1 = p1:xs
          aux (Pattern prod nt rhs:xs) p1 = (Pattern prod nt (aux rhs p1)):xs
          aux (x:xs)                   p1 = x:aux xs p1
          addCanBe []             canBes = canBes
          addCanBe (p@(f,t,_):xs) canBes = let filtered = filter (\(f1,t1,_) -> (f1==f && t1==t) || f==t)  canBes
                                           in  if (length filtered) == 0 then p:addCanBe xs canBes else addCanBe xs canBes

-- -- -- Instantiation of Show for CanBes
instance (RegExp termA, RegExp termB, Show nonTermA, Show nonTermB, Show prodNameA, Show prodNameB)
         => Show (CanBes termA nonTermA prodNameA termB nonTermB prodNameB) where
  show (CanBes l) = unlines $ map aux l
    where aux (nt1, nt2, prod) = "Non terminal "++show nt1++" canBe "++show nt2++" with production -> "++show prod

-- -- -- Auxiliary Functions
isPatternCanBe :: Pattern termA nonTermA prodNameA termB nonTermB prodNameB -> Bool
isPatternCanBe (P_Get _ (P_Arg _ _)) = True
isPatternCanBe _                     = False
















