{-# LANGUAGE FlexibleInstances #-}

module Get2Silver where

import IntermediateTransfType
import Data.List
import Types

get2Silver :: (Ord prodNameA, Ord prodNameB, Ord nonTermA, Ord nonTermB, Ord termA, Ord termB,
	           Show nonTermA, Show nonTermB, Show prodNameB, Show prodNameA, Show termA, Show termB,
	           RegExp termA, RegExp termB)
           => Transf termA nonTermA prodNameA termB nonTermB prodNameB -> String
get2Silver transf = let interm = transf2IntermediateTransf transf
                        header = "grammar Gram;\n"
                        attrs  = writeAttrs transf
                        prods  = writeProds interm
                    in  unlines [header, attrs, prods]

writeAttrs :: (Show nonTermA, Show nonTermB) => Transf termA nonTermA prodNameA termB nonTermB prodNameB -> String
writeAttrs (Transf rules) = unlines . nub . map (aux . fst) $ rules
                          where aux (P_Get to patt) = let from = getPatternType patt
                                                      in  "synthesized attribute get_"++show from++"_"++show to++" :: "++show to++" occurs on "++show from++";"

writeProds :: (Show nonTermA, Show nonTermB, Show prodNameB, Show prodNameA, Show termA, Show termB, RegExp termA, RegExp termB)
           => IntermediateTransf termA nonTermA prodNameA termB nonTermB prodNameB -> String
writeProds (IntermediateTransf inter) = unlines $ map createSilverProd inter

-- All the rules inside this list of lists have exactly the same lhs
-- This list will never be empty, there will always be at least one rule
createSilverProd :: (Show nonTermA, Show nonTermB, Show prodNameB, Show prodNameA, Show termA, Show termB, RegExp termA, RegExp termB)
                 => [[Rule termA nonTermA prodNameA termB nonTermB prodNameB]] -> String
createSilverProd l = let prod = (\(P_Get _ prod) -> prod) . fst . head . head $ l
                         s1   = "aspect production p_" ++ (show $ getPatternProd prod)
                         s2   = signature prod
                         s3   = intercalate "\n\n" $ map createAttrs l
                     in  unlines [s1,s2,"{",s3,"}"]
                     where signature (Pattern prod nt l) = "this::"++show nt++" ::= "++(my_concat ['a'..'z'] (map aux l))
                           aux (P_Term p  )    = "'"++regExp p++"'"
                           aux (P_Prim n t)    = show t
                           aux (P_Arg  n t)    = show t
                           aux (Pattern _ t _) = show t
                           aux (P_Lexemme _ _ t) = show t
                           -- We need my_concat because I don't want to use variable names in production signatures with terminals,
                           -- so I will only create things like: a::A and never things like a::';'
                           my_concat _      []     = ""
                           -- (head y == '\'') is how I know its a terminal :-)
                           my_concat (x:xs) (y:ys) = if (head y == '\'') then y++" "++my_concat (x:xs) ys else [x]++"::"++y++" "++my_concat xs ys

createAttrs :: (Show nonTermA, Show nonTermB, Show prodNameB, Show prodNameA, Show termA, Show termB, RegExp termA, RegExp termB)
            => [Rule termA nonTermA prodNameA termB nonTermB prodNameB] -> String
createAttrs l =  let number_attrs = length l
                     attr_name    = "get_"++from++"_"++to
                     s1           = "  this."++attr_name++" = "++attr_name++"_"++show number_attrs++";"
                     -- The zip is to know which local attribute I am creating when creating each local get
                     s2           = intercalate "\n" . map createAttr $ zip (reverse [1..number_attrs]) l
                     s3           = "  local attribute " ++attr_name++"_0 :: "++to++" = error (\"Attribute "++attr_name++" failed to pattern match on production "++prod++".\");"
                 in  intercalate "\n" [s1,s2,s3]
                 where random_patt = fst . head $ l
                       from        = show . (\(P_Get _  patt) -> getPatternType patt) $ random_patt
                       to          = show . getPatternType $ random_patt
                       prod        = show . (\(P_Get _  patt) -> getPatternProd patt) $ random_patt

createAttr :: (Show nonTermA, Show nonTermB, Show prodNameB, Show prodNameA, Show termA, Show termB, RegExp termA, RegExp termB)
           => (Int, Rule termA nonTermA prodNameA termB nonTermB prodNameB) -> String
createAttr (int, rule) = let s1    = "  local attribute get_"++from++"_"++to++"_"++show int++" :: "++to++" = case this of"
                             ident = spaces ((length s1) - 14) -- 14 is to compensate the static part. Only the name of the attribute changes.
                             s2    = ident ++ "| " ++ sourcePattern2Silver (fst rule) ++ " -> " ++ targetPattern2Silver from (snd rule)
                             s3    = ident ++ "| _ -> " ++ "get_"++from++"_"++to++"_"++show (int-1)
                             s4    = ident ++ "end;"
                         in  intercalate "\n" [s1, s2, s3, s4]
                         where from = show . (\(P_Get _  patt) -> getPatternType patt) . fst $ rule
                               to   = show . getPatternType . fst $ rule
                               prod = show . (\(P_Get _  patt) -> getPatternProd patt) . fst $ rule

sourcePattern2Silver :: (Show nonTermA, Show nonTermB, Show prodNameB, Show prodNameA, Show termA, Show termB, RegExp termA, RegExp termB)
                     => (Pattern termA nonTermA prodNameA termB nonTermB prodNameB) -> String
sourcePattern2Silver (P_Arg   n t    ) = [n]
sourcePattern2Silver (Pattern p t rhs) = "p_"++(show p)++"( "++(intercalate ", " $ map sourcePattern2Silver rhs)++" )"
sourcePattern2Silver (P_Lexemme n p t) = [n]
sourcePattern2Silver (P_Term    p    ) = "_"
sourcePattern2Silver (P_Prim  n t    ) = [n]
sourcePattern2Silver (P_Get t   p    ) = sourcePattern2Silver p

targetPattern2Silver :: (Show nonTermA, Show nonTermB, Show prodNameB, Show prodNameA, Show termA, Show termB, RegExp termA, RegExp termB)
                     => String -> (Pattern termA nonTermA prodNameA termB nonTermB prodNameB) -> String
targetPattern2Silver nonTermS (Pattern p t rhs) = let s1 = map (targetPattern2Silver_aux True) rhs ++ ["link = link"++nonTermS++"(this)"]
                                                  in  "p_"++(show p)++"( "++intercalate ", " s1++" )" 
targetPattern2Silver nonTermS (P_Get   n p    ) = targetPattern2Silver_aux True p++".get_"++show (getPatternType p)++"_"++show n

-- Why targetPattern2Silver_aux need a Boolean as argument? There are two cases for when to insert the links back:
-- 1) The first pattern is always created with the link back to the original tree. This is handled
--    by targetPattern2Silver which always inserts a link.
-- 2) If the target pattern has other target patterns inside, by definition these also need to
--    have links back but they are null, that is why when targetPattern2Silver_aux is called with True 'link=none()' is
--    inserted. One example of cases 1) and 2) is: P_ADD (a, P_ADD(b,c, link=none()), link = linkE(this))
-- 3) When a pattern comes from a production P_Get it is actually a pattern on the source from which
--    attributes still need to be computated. In this cases no link back is needed and targetPattern2Silver_aux is called
--    with False. For example: P_ADD (a, P_Add(b,c).get, link = linkE(this))
targetPattern2Silver_aux :: (Show nonTermA, Show nonTermB, Show prodNameB, Show prodNameA, Show termA, Show termB, RegExp termA, RegExp termB)
    => Bool -> (Pattern termA nonTermA prodNameA termB nonTermB prodNameB) -> String
targetPattern2Silver_aux link (P_Term    p    ) = "terminal ("++show p++", \""++regExp p++"\")"
targetPattern2Silver_aux link (P_Prim  n t    ) = [n]
targetPattern2Silver_aux link (P_Arg   n t    ) = [n]
targetPattern2Silver_aux link (Pattern p t rhs) = let s1 = map (targetPattern2Silver_aux True) rhs
                                                      s2 = if link then s1++["link = none()"] else s1
                                                  in  "p_"++(show p)++"( "++intercalate ", " s2++" )"
targetPattern2Silver_aux link (P_Lexemme n p t) = "terminal ("++show t++", \""++p++"\")"
targetPattern2Silver_aux link (P_Get   n p    ) = targetPattern2Silver_aux False p++".get_"++show (getPatternType p)++"_"++show n















