module IntermediateTransfType where

import Types
import Data.List
import Data.Maybe

-- How do we generate the intermediate transformation datatype to help in generating Silver code:
-- 1) We group the rules by pattern, which will correspond to productions in Silver
-- 2) We group the result of 1) by target non-terminal, which wil correspond to
--    available transformations for production which target differente non terminals
-- 3) We sort the result of 2) per pattern depth, starting on the longest, which will be tried first in Silver
-- 4) We arrange the variables names. Because on pattern signature NT's are always names from 'a' to 'z', we
--    go through an additional step to ensure that on the lhs of a rule variables are also named 'a' to 'z'
--    and arrange the rhs of a rule accordingly.
--    For example (Patt 'c' 'x' 'g' -> Patt 'x' 'c' 'g') becomes (Patt 'a' 'b' 'c' -> Patt 'b' 'a' 'c')
--    This last step really simplificates generating Silver code.

data IntermediateTransf termA nonTermA prodNameA termB nonTermB prodNameB = IntermediateTransf [[[ Rule termA nonTermA prodNameA termB nonTermB prodNameB ]]]
  deriving (Show)

transf2IntermediateTransf :: (Ord termA, Ord nonTermA, Ord prodNameA, Ord termB, Ord nonTermB, Ord prodNameB)
                          => Transf termA nonTermA prodNameA termB nonTermB prodNameB
                          -> IntermediateTransf termA nonTermA prodNameA termB nonTermB prodNameB
transf2IntermediateTransf (Transf t) = let -- step 1) (slightly complicated because for groupBy to work, the list has to be sorted first)
                                            s1 = groupBy sourceProdEq . sortBy sourceProdSort $ t
                                            -- step 2) (slightly complicated because for groupBy to work, the list has to be sorted first)
                                            s2 = map (groupBy targetTypeEq . sortBy targetTypeSort) s1
                                            -- step 3) (reverse because I want them ordered descending, not ascending as sortBy does by default)
                                            s3 = map ({-reverse .-} map (sortBy patternLengthSort)) s2
                                            -- step 4)
                                            s4 = map (map (map arrange_variables_names)) s3
                                        in  IntermediateTransf s4

-- Auxiliary functions for sorting and ordering rules
sourceProdEq :: Eq prodNameB
   => Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Bool
sourceProdEq rule1 rule2 = getPatternProd pat1_1 == getPatternProd pat2_1
                         where pat1_1 = getProd . fst $ rule1
                               pat2_1 = getProd . fst $ rule2
                               getProd (P_Get _ p) = p
                               getProd (P_Put _ p) = p

sourceProdSort :: Ord prodNameB
   => Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Ordering
sourceProdSort rule1 rule2 | getPatternProd pat1_1 == getPatternProd pat2_1 = EQ
                           | getPatternProd pat1_1 >  getPatternProd pat2_1 = GT
                           | otherwise                                      = LT
                           where pat1_1 = getProd . fst $ rule1
                                 pat2_1 = getProd . fst $ rule2
                                 getProd (P_Get _ p) = p
                                 getProd (P_Put _ p) = p

targetTypeEq :: Eq nonTermA
   => Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Bool
targetTypeEq rule1 rule2 = getPatternType pat1_2 == getPatternType pat2_2
                         where pat1_2 = snd rule1
                               pat2_2 = snd rule2

targetTypeSort :: Ord nonTermA
   => Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Ordering
targetTypeSort rule1 rule2 | getPatternType pat1_1 == getPatternType pat2_1 = EQ
                           | getPatternType pat1_1 >  getPatternType pat2_1 = GT
                           | otherwise                                      = LT
                           where pat1_1 = snd rule1
                                 pat2_1 = snd rule2

patternLengthSort :: Ord nonTermA
   => Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Rule termA nonTermA prodNameA termB nonTermB prodNameB
   -> Ordering
patternLengthSort rule1 rule2 | getPatternDepth (fst rule1) > getPatternDepth (fst rule1) = GT
                              | getPatternDepth (fst rule1) < getPatternDepth (fst rule1) = LT
                              | otherwise                                                 = EQ

-- Auxiliary functions for step 4)
arrange_variables_names :: Rule termA nonTermA prodNameA termB nonTermB prodNameB -> Rule termA nonTermA prodNameA termB nonTermB prodNameB
arrange_variables_names (p1,p2) = let -- The first step collects all the vars
                                      -- We need to filter for repetition (nub) because the programmer might 
                                      -- have patterns with variable repetitions, as seen on the Decls example
                                      getVars = zip (nub $ collect p1) ['a'..'z'] :: [(Char, Char)]
                                      -- The second step performs the substitution
                                  in  (change getVars p1, change getVars p2)
                                  where collect :: Pattern termA nonTermA prodNameA termB nonTermB prodNameB -> String
                                        collect (P_Arg     v _   ) = [v]
                                        collect (P_Prim    v _   ) = [v]
                                        collect (P_Lexemme v _ _ ) = [v]
                                        collect (Pattern _ _ rhs ) = concat $ map collect rhs
                                        collect (P_Term  _       ) = []
                                        collect (P_Get   _ p     ) = collect p
                                        collect (P_Put   _ p     ) = collect p
                                        change :: [(Char, Char)] -- [(Char, Char)] is the symbol table
                                               -> Pattern termA nonTermA prodNameA termB nonTermB prodNameB
                                               -> Pattern termA nonTermA prodNameA termB nonTermB prodNameB
                                        change l (P_Arg     v r     ) = P_Arg     (fromJust $ lookup v l) r
                                        change l (P_Prim    v r     ) = P_Prim    (fromJust $ lookup v l) r
                                        change l (P_Lexemme v r1 r2 ) = P_Lexemme v r1 r2
                                        change l (Pattern p t rhs   ) = Pattern   p t (map (change l) rhs) 
                                        change l (P_Term  r         ) = P_Term    r
                                        change l (P_Get   nt patt   ) = P_Get     nt (change l patt) 
                                        change l (P_Put   nt patt   ) = P_Put     nt (change l patt) 

