{-# LANGUAGE TypeSynonymInstances, FlexibleInstances #-}

module GetTransf2Silver where

import Bidi
import Data.List
import Transf2Silver
import Debug.Trace

-- -- ****** This module is EXACLY the same as PutTransf2Silver.hs, but instead of creating a Silver AG with
-- -- an attribute called putTransf, it creates one with a module names getTransf.
-- -- The inherent mechanism is exactly the same and should continue that way.
-- -- Because we will overlapping instances, we had to rename the class toSilver to toSilverGet


-- This Module looks more complicated than it seems, so for each instance of toSilver
-- I will explain in detail what it does.

-- First, we use the function transf2SilverTransf to convert our transformation datatype
-- into something easy to manage. Please refer to the function to see what it does
-- (basically, we sort everything into lists of lists, sorted first by their production,
-- and then by their length).
instance (Ord termA, Ord nonTermA, Ord prod_nameA, Ord termB, Ord nonTermB, Ord prod_nameB,
          Show nonTermB, Show nonTermA, Show prod_nameA, Show prod_nameB, Show termB,
          RegExp termB, Show termA, RegExp termA) =>
         CreateGet (Transf termA nonTermA prod_nameA termB nonTermB prod_nameB) where
  toSilverGet = toSilverGet . transf2SilverTransf

-- This instance initializes the attributes declaration
instance (Show nonTermB, Show nonTermA, Show prod_nameA, Show prod_nameB, Show termB, RegExp termB,
          Show termA, RegExp termA) =>
         CreateGet (IntermediateTransf termA nonTermA prod_nameA termB nonTermB prod_nameB) where
  toSilverGet transf =
          -- by default all grammars are called Gram
      let header = "grammar Gram;\n\n"
	      -- attrs is the attributes declarations
          attrs  = toSilverGet (interTransfNT transf)
		  -- prods are the attributes specifications
          prods  = concat . map toSilverGet $ (orderedRules transf)
      in  concat [header, attrs, prods]

-- This instance handles the attributes declarations
instance (Show nonTermB, Show nonTermA) => CreateGet (NTFunc nonTermA nonTermB) where
  toSilverGet nt = let attributes = unlines . nub . map (\(a,b) -> auxAttrs b) $ nt
                       occurences = unlines . nub . map (\(a,l) -> auxOcc a l) $ nt
                   in  concat [attributes, occurences]
         where auxAttrs = unlines . map (\x -> "synthesized attribute get_"++show x++" :: "++show x++";")
               auxOcc a = concat . map (\x -> "attribute get_"++show x++" occurs on "++show a++";")

-- This instance initializes the productions where the attributes will be specified
-- (the input lists have rules whose lhs productions are exactly the same)
instance (Show prod_nameA, Show nonTermB, Show prod_nameB, Show termB, RegExp termB,
          Show nonTermA, Show termA, RegExp termA) =>
         CreateGet [[Rule termA nonTermA prod_nameA termB nonTermB prod_nameB]] where
  toSilverGet l =
      let a1       = "aspect production p_" ++ (show . getPatternProd . fst . head . head $ l)
-- **** VERY IMPORTANTE: for this function to work, it is assumed there ALWAYS is a default
-- rule in the transformation, i.e., a rule that maps directly to the production's
-- children, with no need to pattern match. This rule is used to define the abstract
-- production in Silver, such as: this::A ::= a::A b::A
          a2       = (\(Pattern (prod,nt) l) -> "this::"++show nt++" ::= "++(intercalate " " $ map aux l)) . fst . last . head $ l
          listVars = (\(Pattern (prod,nt) l) -> getListVars l) . fst . last . head $ l
          a3       = concat . map toSilverGet $ map (\x -> (x,listVars)) l
      in  concat [a1,"\n",a2,"\n","{","\n",a3,"\n","}","\n\n"]
      where aux (P_Term p)          = show p
            aux (P_Prim (n,t))      = [n]++"::"++show t
            aux (P_Arg  (n,t))      = [n]++"::"++show t
			-- For why this gives an error, check the comment below
            aux (Pattern (n,t) rhs) = error "*** For this error, refer to Transf2Silver.hs, line 59 ***"
	  

-- This instance initializes the attributes per non-terminal
-- and calls instances to create each local attribute
-- The "error" attribute is also created here
instance (Show prod_nameA, Show nonTermB, Show prod_nameB, Show termB, RegExp termB,
          Show termA, Show nonTermA, RegExp termA) =>
         CreateGet ([Rule termA nonTermA prod_nameA termB nonTermB prod_nameB], [Char]) where
  toSilverGet (l,listVars)
     =  let lengthR = length l
            nonTerm = show . getPatternType . snd . head $ l
            a1      = "  this.get_"++nonTerm++" = get_"++nonTerm++"_"++show lengthR++";"
            -- The zip is to know which local attribute I am creating when 
            -- implementing 'map toSilver'
            a2      = concat . map toSilverGet $ addList (zip (reverse [1..lengthR]) l) listVars
            a3      = "  local attribute get_"++nonTerm++"_0 :: "++nonTerm++" = error (\"This transformation is not total for non terminal "++nonTerm++"\");"
         in  concat [a1,"\n",a2,"\n",a3]

-- This instance creates a list of 
instance (Show prod_nameB, Show termB, RegExp termB, Show nonTermB,
          Show termA, Show nonTermA, Show prod_nameA, RegExp termA) =>
         CreateGet (Int, Rule termA nonTermA prod_nameA termB nonTermB prod_nameB, [Char]) where
  toSilverGet (n,r, listVars)
     = let nonTerm       = show . getPatternType . snd $ r
           listlexes     = getListLexemes (snd r)
           a1            = "  local attribute get_"++nonTerm++"_"++show n++" :: "++nonTerm++" = "
           -- matchingPatt creates the cases to pattern match subtree.
           -- The arguments are: pattern to be matched (fst r), the pattern
           -- it is transformed into, in case it matches, in string form,
           -- and the local attr the case goes to in case the matching 
           -- fails. The last int is only for identation
           matchingPatt  = pattern2Cases (fst r) resultingPatt ("get_"++nonTerm++"_"++show (n-1)) 27 listVars
           resultingPatt = printTargetPattern (getPatternType $ fst r) (snd r)
       in  a1++matchingPatt

-- This function takes the target pattern and transforms it into a the resulting
-- subtree, and creates the recursive calls to this tree non-terminals
printTargetPattern :: (RegExp term, Show term, Show nonTerm, Show prodName, Show nonTermS) => nonTermS -> (Bidi.Pattern term nonTerm prodName) -> String
printTargetPattern nonTermS (P_Term  p   )        = "terminal ("++show p++", \""++regExp p++"\" )"
printTargetPattern nonTermS (P_Prim (n,t))        = [n]
printTargetPattern nonTermS (P_Arg  (n,t))        = [n]++".get_"++(show t)
printTargetPattern nonTermS (Pattern (p,t) rhs)   = "p_"++(show p)++"( "++(intercalate ", " $ map auxprintTargetPattern rhs)++", link = link"++show nonTermS++"(this) )"
printTargetPattern nonTermS (P_Lexemme (n,p,t)  ) = "terminal ("++show t++", \""++p++"\")"

auxprintTargetPattern (P_Term  p   )        = "terminal ("++show p++", \""++regExp p++"\")"
auxprintTargetPattern (P_Prim (n,t))        = [n]
auxprintTargetPattern (P_Arg  (n,t))        = [n]++".get_"++(show t)
auxprintTargetPattern (Pattern (p,t) rhs)   = "p_"++(show p)++"( "++(intercalate ", " $ map auxprintTargetPattern rhs)++", link = none())"
auxprintTargetPattern (P_Lexemme (n,p,t)  ) = "terminal ("++show t++", \""++p++"\")"
 


