{-# LANGUAGE TypeSynonymInstances, FlexibleInstances #-}

module PutTransf2Silver where

import Bidi
import Data.List
import Transf2Silver
import Debug.Trace

-- -- ****** This module is EXACLY the same as GetTransf2Silver, but instead of creating a Silver AG with
-- -- an attribute called getTransf, it creates one with a module names putTransf.
-- -- The inherent mechanism is exactly the same and should continue that way.
-- -- Because we will overlapping instances, we had to rename the class toSilver to toSilverPut


-- This Module looks more complicated than it seems, so for each instance of toSilver
-- I will explain in detail what it does.

-- First, we use the function transf2SilverTransf to convert our transformation datatype
-- into something easy to manage. Please refer to the function to see what it does
-- (basically, we sort everything into lists of lists, sorted first by their production,
-- and then by their length).
instance (Ord termA, Ord nonTermA, Ord prod_nameA, Ord termB, Ord nonTermB, Ord prod_nameB,
          Show nonTermB, Show nonTermA, Show prod_nameA, Show prod_nameB, Show termB,
          RegExp termB, Show termA, RegExp termA) =>
--         CreatePut (Transf termA nonTermA prod_nameA termB nonTermB prod_nameB, Transf termB nonTermB prod_nameB termA nonTermA prod_nameA) where
           CreatePut (Transf termA nonTermA prod_nameA termB nonTermB prod_nameB, Transf termB nonTermB prod_nameB termA nonTermA prod_nameA) where
  -- get will be used by the function links_management (we need the CanBes), IntermediateTransf, resulting from
  -- 'transf2SilverTransf put' will be used for anything else (generate Silver code)
  toSilverPut (put,get) = toSilverPut (get, transf2SilverTransf put)

-- This instance initializes the attributes declaration
instance (Show nonTermB, Show nonTermA, Show prod_nameA, Show prod_nameB, Show termB, RegExp termB,
          Show termA, RegExp termA, Eq nonTermB, Ord nonTermA, Ord nonTermB) =>
         CreatePut (Transf termB nonTermB prod_nameB termA nonTermA prod_nameA, IntermediateTransf termA nonTermA prod_nameA termB nonTermB prod_nameB) where
  toSilverPut (transf, intermTransf) =
          -- by default all grammars are called Gram
      let header = "grammar Gram;\n\n"
        -- attrs is the attributes declarations
          attrs  = toSilverPut (interTransfNT intermTransf)
      -- prods are the attributes specifications
          prods  = concat . map (\x -> toSilverPut (transf,x))$ (orderedRules intermTransf)
      in  concat [header,attrs, prods]

-- This instance handles the attributes declarations
instance (Show nonTermB, Show nonTermA) => CreatePut (NTFunc nonTermA nonTermB) where
  toSilverPut nt = let attributes = unlines . nub . map (\(a,b) -> auxAttrs b) $ nt
                       occurences = unlines . nub . map (\(a,l) -> auxOcc a l) $ nt
                   in  concat [attributes, occurences]
         where auxAttrs = unlines . map (\x -> "synthesized attribute put_"++show x++" :: "++show x++";")
               auxOcc a = unlines . map (\x -> "attribute put_"++show x++" occurs on "++show a++";")

-- This instance initializes the productions where the attributes will be specified
-- (the input lists have rules whose lhs productions are exactly the same)
instance (Show prod_nameA, Show nonTermB, Show prod_nameB, Show termB, RegExp termB,
          Show nonTermA, Show termA, RegExp termA, Eq nonTermB, Ord nonTermA, Ord nonTermB) =>
         CreatePut (Transf termB nonTermB prod_nameB termA nonTermA prod_nameA,[[Rule termA nonTermA prod_nameA termB nonTermB prod_nameB]]) where
  toSilverPut (transf,l) =
      let a1       = "aspect production p_" ++ (show . getPatternProd . fst . head . head $ l)
-- **** VERY IMPORTANTE: for this function to work, it is assumed there ALWAYS is a default
-- rule in the transformation, i.e., a rule that maps directly to the production's
-- children, with no need to pattern match. This rule is used to define the abstract
-- production in Silver, such as: this::A ::= a::A b::A
          a2       = (\(Pattern (prod,nt) l) -> "this::"++show nt++" ::= "++(intercalate " " $ map aux l)) . fst . last . head $ l
          listVars = (\(Pattern (prod,nt) l) -> getListVars l) . fst . last . head $ l
          a3      = concat . map toSilverPut $ map (\x -> (transf, x, listVars)) l
      in  concat [a1,"\n",a2,"\n","{","\n",a3,"\n","}","\n\n"]
      where aux (P_Term p)          = show p
            aux (P_Prim (n,t))      = [n]++"::"++show t
            aux (P_Arg  (n,t))      = [n]++"::"++show t
      -- For why this gives an error, check the comment above
            aux (Pattern (n,t) rhs) = error "*** For this error, refer to Transf2Silver.hs, line 59 ***"
    

-- This instance initializes the attributes per non-terminal
-- and calls instances to create each local attribute
-- The "error" attribute is also created here
instance (Show prod_nameA, Show nonTermB, Show prod_nameB, Show termB, RegExp termB,
          Show termA, Show nonTermA, RegExp termA, Eq nonTermB, Ord nonTermA, Ord nonTermB) =>
         CreatePut (Transf termB nonTermB prod_nameB termA nonTermA prod_nameA, [Rule termA nonTermA prod_nameA termB nonTermB prod_nameB], [Char]) where
  toSilverPut (transf, l, listVars)
     =  let lengthR = length l
            nonTerm = getPatternType . snd . head $ l
            a11     = "  this.put_"++show nonTerm++" = "
            a1      = a11 ++ links_management (length a11) nonTerm transf ("put_"++show nonTerm++"_"++show lengthR) ++"\n"++spaces 15++"end;"
            -- The zip is to know which local attribute I am creating when 
            -- implementing 'map toSilver'
            a2      = concat . map toSilverPut $ addList (zip (reverse [1..lengthR]) l) listVars
            a3      = "  local attribute put_"++show nonTerm++"_0 :: "++show nonTerm++" = error (\"This transformation is not total for non terminal "++show nonTerm++"\");\n\n"
         in concat [a1,"\n",a2,"\n",a3]

-- This instance creates a list of 
instance (Show prod_nameB, Show termB, RegExp termB, Show nonTermB,
          Show termA, Show nonTermA, Show prod_nameA, RegExp termA) =>
         CreatePut (Int, Rule termA nonTermA prod_nameA termB nonTermB prod_nameB, [Char]) where
  toSilverPut (n,r, listVars)
     = let nonTerm       = show . getPatternType . snd $ r
           listlexes     = getListLexemes (snd r)
           a1            = "  local attribute put_"++nonTerm++"_"++show n++" :: "++nonTerm++" = "
           -- matchingPatt creates the cases to pattern match subtree.
           -- The arguments are: pattern to be matched (fst r), the pattern
           -- it is transformed into, in case it matches, in string form,
           -- and the local attr the case goes to in case the matching 
           -- fails. The last int is only for identation
           matchingPatt  = pattern2Cases (fst r) resultingPatt ("put_"++nonTerm++"_"++show (n-1)) 27 listVars
           resultingPatt = printTargetPattern (snd r)
       in  a1++matchingPatt

-- This function takes the target pattern and transformas it into a the resulting
-- subtree, and creates the recursive calls to this tree non-terminals
printTargetPattern :: (RegExp term, Show term, Show nonTerm, Show prodName) => (Bidi.Pattern term nonTerm prodName) -> String
printTargetPattern (P_Term  p   )        = "terminal ("++show p++", \""++regExp p++"\")"
printTargetPattern (P_Prim (n,t))        = [n]
printTargetPattern (P_Arg  (n,t))        = [n]++".put_"++(show t)
printTargetPattern (Pattern (p,t) rhs)   = "p_"++(show p)++"( "++(intercalate ", " $ map printTargetPattern rhs)++" )"
printTargetPattern (P_Lexemme (n,p,t)  ) = "terminal ("++show t++", \""++p++"\")"

---- The Int is for identation
--links_management :: (Eq nonTermB, Eq nonTermA, Ord nonTermA, Ord nonTermB, Show prod_nameA, Show nonTermA, Show termA, Show prod_nameB, Show nonTermB, Show termB)
--                 => Int -> nonTermA -> Transf termA nonTermA prod_nameA termB nonTermB prod_nameB -> String -> String
links_management :: (Eq nonTermB, Ord nonTermA, Ord nonTermB, Show termB, Show nonTermB, Show prod_nameB, Show termA, Show nonTermA, Show prod_nameA, RegExp termA)
                 => Int -> nonTermA -> Transf termA nonTermA prod_nameA termB nonTermB prod_nameB -> String -> String
links_management ident nonTerm transf default_case =
  let nonTerms = clean_nt $ transfNT transf
      begin    = "case this.link of"
      cases    = concat $ map (create_cases ident nonTerm transf) nonTerms
      end      = spaces ident ++ "| none() -> " ++ default_case
  in  concat [begin,cases,"\n",end]  
  where clean_nt nt = nub  $ map (\(l,_) -> l) nt

create_cases :: (Eq nonTermA, Ord nonTermB, Ord nonTermA, Show prod_nameA, Show termA, Show prod_nameB, Show nonTermB, Show termB, Show nonTermA, RegExp termA)
             => Int -> nonTermA -> Transf termA nonTermA prod_nameA termB nonTermB prod_nameB -> nonTermA -> String
create_cases ident targetNonTerm transf linkNonTerm = let s1 = "\n"++spaces ident++"| link"++show linkNonTerm++"(e) -> "
                                                          s2 = if (targetNonTerm==linkNonTerm) then "e" else manage_canbes targetNonTerm linkNonTerm transf
                                                      in  concat [s1,s2]

manage_canbes :: (Eq nonTermA, Ord nonTermB, Ord nonTermA, Show prod_nameA, Show nonTermA, Show termA, Show prod_nameB, Show nonTermB, Show termB, RegExp termA)
              => nonTermA -> nonTermA -> Transf termA nonTermA prod_nameA termB nonTermB prod_nameB -> String
manage_canbes targetNonTerm linkNonTerm (Transf nt rules) = let inverted_nt     = invert_nt nt
                                                                inverted_rules  = transf2TransfAux inverted_nt rules
                                                                canBes          = if (length inverted_rules == 0) then [] else canbes . head $ inverted_rules
                                                                rule            = head $ findCanBe targetNonTerm linkNonTerm canBes canBes
                                                            in  apply_rule rule

--- Auxiliary Functions to deal with CanBes ---
findCanBe :: Eq nonTermB => nonTermB -> nonTermB
          -> [CanBe termA nonTermA prod_nameA termB nonTermB prod_nameB]
          -> [CanBe termA nonTermA prod_nameA termB nonTermB prod_nameB]
          -> [CanBe termA nonTermA prod_nameA termB nonTermB prod_nameB]
findCanBe targetNonTerm linkNonTerm canBes extendedCanBes =
  let s1 = filter (\x -> (canBetoType x == targetNonTerm && canBefromType x == linkNonTerm)) extendedCanBes
  in  if (length s1 > 0) then s1 else findCanBe targetNonTerm linkNonTerm canBes (extendCanBes canBes extendedCanBes)

extendCanBes :: Eq nonTermB
     => [CanBe termA nonTermA prod_nameA termB nonTermB prodNameB]
     -> [CanBe termA nonTermA prod_nameA termB nonTermB prodNameB]
     -> [CanBe termA nonTermA prod_nameA termB nonTermB prodNameB]
extendCanBes canBes extendedCanBes = map (search canBes) extendedCanBes
  where search :: (Eq nonTermB)
               => [CanBe termA nonTermA prod_nameA termB nonTermB prodNameB]
               -> CanBe termA nonTermA prod_nameA termB nonTermB prodNameB
               -> CanBe termA nonTermA prod_nameA termB nonTermB prodNameB
        search canBes canBe = let s1 = filter (\x -> canBefromType canBe == canBetoType x) canBes
                              -- Once again, we assume only one solution is found
                              in apply (head s1) canBe
        apply :: CanBe termA nonTermA prod_nameA termB nonTermB prodNameB -> CanBe termA nonTermA prod_nameA termB nonTermB prodNameB
              -> CanBe termA nonTermA prod_nameA termB nonTermB prodNameB
        apply canBe (p1, Pattern pair rhs) = let s1 = map (aux canBe) rhs
                                             in  (p1, Pattern pair s1)
                                             where aux canBe (P_Arg     p    ) = snd canBe
                                                   aux canBe (P_Term    p    ) = (P_Term    p)
                                                   aux canBe (P_Lexemme p    ) = (P_Lexemme p)
                                                   aux canBe (Pattern   p rhs) = (Pattern   p (map (aux canBe) rhs))

apply_rule :: (Show termB, Show prodNameB, RegExp termB) => CanBe termA nonTermA prod_nameA termB nonTermB prodNameB -> String
apply_rule (p1, Pattern (patt, nonTerm) rhs) = "p_"++show patt++"("++(intercalate ", " $ map aux rhs)++")"
                                             where aux :: (Show prodName, RegExp term, Show term) => Pattern term nonTerm prodName -> String
                                                   aux (P_Arg     p                  ) = "e"
                                                   aux (P_Term    p                  ) = "terminal ("++show p++", \""++regExp p++"\")"
                                                   aux (P_Lexemme p                  ) = ""
                                                   aux (Pattern   (patt, nonTerm) rhs) = "p_"++show patt++"("++(intercalate ", " $ map aux rhs)++")"












