module Transf2Silver where 

import Bidi
import Data.List
import Data.Maybe

-- **  This module has functions and datatypes that are common to both PutTransf2Silver and GetTransf2Silver ** --


-- This datatype has exactly the same information as the Transf datatype, but
-- list of rules was preprocessed by transf2SilverTransf
data IntermediateTransf termA nonTermA prod_nameA termB nonTermB prod_nameB
  = IntermediateTransf { interTransfNT     :: NTFunc nonTermA nonTermB,
                         orderedRules      :: [[[ Rule termA nonTermA prod_nameA termB nonTermB prod_nameB ]]] }
                       deriving (Show, Eq, Ord)

-- Function that translates Transf to IntermediateTransf
-- In the end of this function we will end with:
-- [ list of rules [list of rules by their production [list of rules by their target nonterminal, sorted by length]] ]
transf2SilverTransf :: (Ord termA, Ord nonTermA, Ord prod_nameA, Ord termB, Ord nonTermB, Ord prod_nameB)
                    => (Transf termA nonTermA prod_nameA termB nonTermB prod_nameB)
                    -> (IntermediateTransf termA nonTermA prod_nameA termB nonTermB prod_nameB)
transf2SilverTransf t = let nt    = transfNT t
                            -- In this first step, we group the rules by their
              -- lhs pattern prod. For groupBy to work, the list
              -- must be sorted first.
                            rules = groupBy lhsGroup . sortBy lhsSort . transfRules $ t
                            -- Same thing as he previous step, only this time we
                            -- group rules by their rhs pattern type (remember,
                            -- a pattern type is the nonterminal on topmost)
                            -- The last step is sorting rules by the size of their
              -- lhs pattern
                            rulesByNT = map (map (sortBy lengthSort)) . map (groupBy rhsGroup) . map (sortBy rhsSort) $ rules
                        in  IntermediateTransf {interTransfNT = nt, orderedRules = arrange_variables $ rulesByNT}
  where lhsSort a b | getPatternProd (fst a) > getPatternProd (fst b)  = GT
                    | getPatternProd (fst a) < getPatternProd (fst b)  = LT
                    | getPatternProd (fst a) == getPatternProd (fst b) = EQ
        rhsSort a b | getPatternType (snd a) > getPatternType (snd b)  = GT
                    | getPatternType (snd a) < getPatternType (snd b)  = LT
                    | getPatternType (snd a) == getPatternType (snd b) = EQ
        lengthSort a b | getPatternDepth (fst a) < getPatternDepth (fst b) = GT
                       | getPatternDepth (fst a) > getPatternDepth (fst b) = LT
                       | otherwise                                         = EQ
        lhsGroup a b = getPatternProd (fst a) == getPatternProd (fst b)
        rhsGroup a b = getPatternType (snd a) == getPatternType (snd b)


-- We need this next function, arrange_variables, for cases like these:
--     [[(Pattern (While,Stat) [P_Term WHILE,P_Arg ('a',Expr),P_Term LeftBracket,P_Arg ('b',ListStat),P_Term RightBracket],
--       Pattern (While,Stat) [P_Term WHILE,P_Arg ('a',Expr),P_Term LeftBracket,P_Arg ('b',ListStat),P_Term RightBracket])
--     ,(Pattern (While,Stat) [P_Term WHILE,P_Arg ('b',Expr),P_Term LeftBracket,P_Arg ('a',ListStat),P_Term RightBracket],
--       Pattern (RepeatUntil,Stat) [P_Term REPEAT,P_Term LeftBracket,P_Arg ('a',ListStat),P_Term RightBracket,P_Term UNTIL,P_Arg ('b',Expr)])]]
-- Which were generating the wrong Silver code:
--      aspect production p_While
--      this::Stat ::= WHILE b::Expr LeftBracket a::ListStat RightBracket
--      { this.put_Stat = put_Stat_2;
--        local attribute put_Stat_2 :: Stat = case b, a of
--                                   | _, _ -> p_While( terminal (WHILE, "WHILE"), a.put_Expr, terminal (LeftBracket, "{"), b.put_ListStat, terminal (RightBracket, "}") )
--                                 end;
--        local attribute put_Stat_1 :: Stat = case b, a of
--                                    | _, _ -> p_RepeatUntil( terminal (REPEAT, "REPEAT"), terminal (LeftBracket, "{"), a.put_ListStat, terminal (RightBracket, "}"), terminal (UNTIL, "UNTIL"), b.put_Expr )
--                                 end;
--        local attribute put_Stat_0 :: Stat = error ("This transformation is not total for non terminal Stat");}
arrange_variables :: [[[ Rule termA nonTermA prod_nameA termB nonTermB prod_nameB ]]] -> [[[ Rule termA nonTermA prod_nameA termB nonTermB prod_nameB ]]]
arrange_variables list =  map (map (map arrange_rules)) list
arrange_rules :: Rule termA nonTermA prod_nameA termB nonTermB prod_nameB -> Rule termA nonTermA prod_nameA termB nonTermB prod_nameB
                         -- The first step goes through the first Pattern of a rule and
                         -- collects all the vars, and intercalates them with a list ['a'..'z']
                         -- This list will be used to ensure all the variable of the first rule are "well written",
                         -- which means the first one is 'a', the second one 'b', etc
arrange_rules (p1,p2) = let getVars = zip (filter (/= '_') $ aux p1) ['a'..'z'] :: [(Char, Char)]
                         -- The second step goes through both Patterns of the rule
                         -- Variables change but the relation to the second part holds.
                         -- For example, (Pattern 'c' 'a', Pattern 'a' 'c')
                         -- becomes (Pattern 'a' 'b', Pattern 'b' 'a')
                     in  (change getVars p1, change getVars p2)
                     where aux (P_Arg     (v,_)     ) = [v]
                           aux (P_Prim    (v,_)     ) = [v]
                           aux (P_Lexemme (v,_,_)   ) = [v]
                           aux (Pattern   _     rhs ) = concat $ map aux rhs
                           aux (P_Term  _           ) = ['_']
                           change l (P_Arg     (v,r)     ) = (P_Arg  (fromJust $ lookup v l,r) )
                           change l (P_Prim    (v,r)     ) = (P_Prim (fromJust $ lookup v l,r) )
                           change l (P_Lexemme (v,r1,r2) ) = (P_Lexemme (fromJust $ lookup v l,r1,r2) )
                           change l (Pattern   r     rhs ) = (Pattern r (map (change l) rhs) )
                           change l (P_Term    r         ) = (P_Term r)


-- In the instance: CreatePut (or CreateGet) (Int, Rule termA nonTermA prod_nameA termB nonTermB prod_nameB)
-- the arguments of this function are explained.
-- listVars is generated by getListVars
pattern2Cases :: (Show prodName, Show nonTerm, Show term, RegExp term) =>
                 Pattern term nonTerm prodName -> String -> String -> Int -> [Char] -> String
pattern2Cases patt@(Pattern (p,t) rhs) g b ident listVars =
  let childs       = concat (map getKids rhs)
      numberChilds = length childs
      a1           = "case "++(intercalate ", " $ map (\x -> [x]) $ take numberChilds listVars)++" of"
      a2           = spaces ident++"   | "++intercalate ", " (map printSourcePattern childs)++" -> "++drawIfs (getLexemmes patt) g b
    -- We test if all the childs are P_Args here because, it that is true, Silver will
    -- give an "Pattern has overlapping cases" error, so if there are only P_Args, we will
    -- not need a default case with underscores. Nevertheless, we keep the loal attribute
    -- that generates an error because it might be needed for other cases.
      a3           = if (areAllChildsP_Args rhs) then ""
                     else "\n"++spaces ident++"   | "++ (intercalate ", " (under numberChilds)) ++ " -> " ++ b
      a4           = spaces ident++"end;"
  in  if (numberChilds == 0) then g++";" else concat [a1,"\n",a2++a3,"\n",a4]
  where getKids     (P_Term  p   )        = []
        getKids     (P_Prim (n,t))        = []
        getKids     (P_Arg  (n,t))        = [P_Arg  (n,t)]
        getKids     (Pattern (p,t) rhs)   = [Pattern (p,t) rhs]
        getKids     (P_Lexemme (n,p,t)  ) = []
        getLexemmes (P_Term  p   )        = []
        getLexemmes (P_Prim (n,t))        = []
        getLexemmes (P_Arg  (n,t))        = []
        getLexemmes (Pattern (p,t) rhs)   = concat (map getLexemmes rhs)
        getLexemmes (P_Lexemme (n,p,t)  ) = [(n,p)]
        drawIfs [] g _ = g
        drawIfs l  g b = let lexemesTests = concat $ map (\(a,b) -> [a]++".lexeme == \""++b++"\"") l
                         in  "if ("++lexemesTests++") then "++g++" else "++b
        areAllChildsP_Args [] = True
        areAllChildsP_Args (P_Term  p         :xs) = True  && areAllChildsP_Args xs
        areAllChildsP_Args (P_Prim (n,t)      :xs) = True  && areAllChildsP_Args xs
        areAllChildsP_Args (P_Arg  (n,t)      :xs) = True  && areAllChildsP_Args xs
        areAllChildsP_Args (Pattern (p,t) rhs :xs) = False && areAllChildsP_Args xs
        areAllChildsP_Args (P_Lexemme (n,p,t) :xs) = False && areAllChildsP_Args xs


-- This function gets a Pattern and creates a symbol table composed
-- by the P_Lexeme's, and the value the terminal must have
-- We are talking about the 'Lexeme' terminals here, which we have
-- to test with if then else (cases do not work).
getListLexemes :: Pattern term nonTerm prod_name -> [(Char,String)]
getListLexemes (P_Term    p        ) = []
getListLexemes (P_Prim    (n,t)    ) = []
getListLexemes (P_Arg     (n,t)    ) = []
getListLexemes (Pattern   (p,t) rhs) = concat . map getListLexemes $ rhs
getListLexemes (P_Lexemme (n,p,t)  ) = [(n,p)]

-- This function takes the target pattern and transformas it into a the resulting
-- subtree, and creates the recursive calls to this tree non-terminals
printSourcePattern :: (RegExp term, Show term, Show nonTerm, Show prodName) => (Bidi.Pattern term nonTerm prodName) -> String
printSourcePattern (P_Arg  (n,t))        = "_"
printSourcePattern (Pattern (p,t) rhs)   = "p_"++(show p)++"( "++(intercalate ", " $ map printSourcePattern rhs)++" )"
printSourcePattern (P_Lexemme (n,p,t)  ) = [n]
printSourcePattern (P_Term  p   )        = "_"
--printSourcePattern (P_Prim (n,t))        = [n]

-- This function gets a pattern and extracts a list with all its P_Arg variables, in the
-- right order
getListVars :: [Pattern t t1 t2] -> [Char]
getListVars [] = ""
getListVars (P_Term p    :xs)      = getListVars xs
getListVars (P_Prim (n,t):xs)      = n : getListVars xs
getListVars (P_Arg  (n,t):xs)      = n : getListVars xs
-- For why this gives an error, check the comment below
-- getListVars (Pattern (n,t) rhs) = error "*** For this error, refer to Transf2Silver.hs, line 59 ***"

-- Aux funcs
under n = take n aux
   where aux = "_":aux
spaces n = take n aux
   where aux = ' ':aux
addList :: [(a,b)] -> [c] -> [(a,b,[c])]
addList [] _ = []
addList ((a,b):xs) l = (a,b,l) : addList xs l

