{-# LANGUAGE DeriveDataTypeable#-}

module Bidi where

import Data.List
import Data.Data

-- ****************************************************************************************--
-- In this module we find:
-- > The definition of the datatypes for Transformation, Primitives and Context-free Grammar
-- > Auxiliary functions for these datatypes
-- > One class to show (on the literal meaning) Primitives in Silver (RegExp)
-- > The implementation of the bidirectionalization engine
-- ****************************************************************************************--


-- All Primitives must be a member of this class, where the information
-- about their regular expression in Silver must be written
-- All the Terminals must also be a member of this class, where their
-- textual information is shown
class RegExp a where
	regExp :: a -> String

-- -- -- Prims are orthogonal to any CFG definition
data Prim = Int_Lit | String_Lit
           deriving (Eq, Show, Ord)

instance RegExp Prim where
  regExp Int_Lit    = "/[0-9]+/"
  regExp String_Lit = "/[a-zA-Z][a-zA-Z0-9_]*/"
  regExp _          = error "A regExp for this literal is not defined. Check Bidi.hs for more information."

-- -- -- We need this because from the concrete view to the abstract view
-- -- -- of the source grammar we loose literals and we transform them into
-- -- -- a new production with a String. For example. terminal (Int_Lit, "3")
-- -- -- becomes int_NT ("3"). We do this so we can pattern match them easily.
-- -- -- These new productions for literals are defined in Terminals.sv
-- -- -- The function literal2Prod just maps literals to the respective productions
-- -- -- The function literal2Type just maps literals to their new type

literal2Prod Int_Lit = "int_NT"
literal2Prod _       = error "A production for this literal is not defined. Check Bidi.hs for more information."

literal2Type Int_Lit = "Int_NT"
literal2Type _       = error "A Type for this literal is not defined. Check Bidi.hs for more information."

-- -- -- Context-free Grammar datatype
data CFG term nonTerm prod_name = CFG { cfgNonTerms  :: [nonTerm]
                                      , cfgPrims     :: [Prim]
                                      , cfgTerms     :: [term]
                                      , cfgProdNames :: [prod_name]
                                      , cfgProds     :: [Prod prod_name nonTerm term]
                                      } deriving (Show, Eq, Ord)

data Prod prod_name nonTerm term = Prod { prodName     :: prod_name
                                        , prodNonTerm  :: nonTerm
                                        , prodRhs      :: [RHS nonTerm term]
                                        } deriving (Show, Eq, Ord)

data RHS nonTerm term = Term term | NonTerm nonTerm | Prim Prim
                      deriving (Show, Eq, Ord)

-- -- -- Transformation datatypes
data Pattern term nonTerm prodName = P_Term    term
                                   | P_Prim    (Char  , Prim)
                                   | P_Arg     (Char  , nonTerm)
                                   | P_Lexemme (Char, String, Prim) -- For example, P_Lexeme('i', "2", Int_Lit)
                                   | Pattern   (prodName,nonTerm) [Pattern term nonTerm prodName]
                                   deriving (Show, Eq, Ord)

type NTFunc nonTermA nonTermB = [(nonTermA,[nonTermB])]

type Rule termA nonTermA prod_nameA termB nonTermB prod_nameB = ( Pattern termA nonTermA prod_nameA
                                                                , Pattern termB nonTermB prod_nameB )

type CanBe termA nonTermA prod_nameA termB nonTermB prod_nameB = ( Pattern termA nonTermA prod_nameA
                                                                 , Pattern termB nonTermB prod_nameB )

data Transf termA nonTermA prod_nameA termB nonTermB prod_nameB = Transf { transfNT :: NTFunc nonTermA nonTermB
                                                                         , transfRules :: [ Rule termA nonTermA prod_nameA 
                                                                                            termB nonTermB prod_nameB ]
                                                                         } deriving (Show, Eq, Ord)

-- -- CanBe auxiliary functions
canBetoType   :: CanBe termA nonTermA prod_nameA termB nonTermB prod_nameB -> nonTermB
canBetoType   = getPatternType . snd

canBefromType :: CanBe termA nonTermA prod_nameA termB nonTermB prod_nameB -> nonTermB
canBefromType = aux . (\(Pattern _ rhs) -> rhs) . snd
              where aux (P_Arg p       : xs) = getPatternType (P_Arg p)
                    aux (P_Term  _     : xs) = aux xs
                    aux (P_Lexemme _   : xs) = aux xs
                    aux (Pattern _ rhs : xs) = aux (rhs++xs)


-- -- -- Auxiliary functions for Patterns
getPatternVar :: Pattern termA nonTermA prod_nameA -> Char
getPatternVar (Pattern  _    _   ) = error "Pattern have no variables!"
getPatternVar (P_Arg   (v,_)     ) = v
getPatternVar (P_Prim  (v,_)     ) = v
getPatternVar (P_Term  _         ) = error "P_Term have no variables!"
getPatternVar (P_Lexemme (v,_,_) ) = v

getPatternType :: Pattern termA nonTermA prod_nameA -> nonTermA
getPatternType (Pattern (_,t) rhs) = t
getPatternType (P_Arg   (_,t) )    = t
-- Prims DO have types. They're not non-terminals though, so we just generate an error.
getPatternType (P_Prim  (_,t) )    = error "Someone, somewhere is calling the function getPatternType on a P_Prim!"
getPatternType (P_Term  term  )    = error "Someone, somewhere is calling the function getPatternType on a P_Term!"
getPatternType (P_Lexemme _     )  = error "Someone, somewhere is calling the function getPatternType on a P_Lexemme!"

getPatternProd :: Pattern termA nonTermA prod_nameA -> prod_nameA
getPatternProd (Pattern (p,_) _)  = p
getPatternProd (P_Arg   (_,t) )   = error "Someone, somewhere is calling the function getPatternProd on a P_Arg!"
getPatternProd (P_Prim  (_,t) )   = error "Someone, somewhere is calling the function getPatternProd on a P_Prim!"
getPatternProd (P_Term  term  )   = error "Someone, somewhere is calling the function getPatternProd on a P_Term!"
getPatternProd (P_Lexemme _     ) = error "Someone, somewhere is calling the function getPatternProd on a P_Lexemme!"

getPatternRHS :: Pattern termA nonTermA prod_nameA -> [Pattern termA nonTermA prod_nameA]
getPatternRHS (Pattern (_,_) rhs) = rhs
getPatternRHS (P_Arg   (_,t)    ) = error "Someone, somewhere is calling the function getPatternRHS on a P_Arg!"
getPatternRHS (P_Prim  (_,t)    ) = error "Someone, somewhere is calling the function getPatternRHS on a P_Prim!"
getPatternRHS (P_Term  term     ) = error "Someone, somewhere is calling the function getPatternRHS on a P_Term!"
getPatternRHS (P_Lexemme _     )  = error "Someone, somewhere is calling the function getPatternRHS on a P_Lexemme!"

getPatternDepth :: Pattern termA nonTermA prod_nameA -> Int
getPatternDepth (Pattern (_,_) rhs) = 1 + (maximum (map getPatternDepth rhs))
getPatternDepth (P_Arg   (_,t)    ) = 1
getPatternDepth (P_Prim  (_,t)    ) = 1
getPatternDepth (P_Term  term     ) = 1
getPatternDepth (P_Lexemme _      ) = 1

-- -- -- Inversion Engine
-- Inverting the nt relation
invert_nt :: (Eq nonTermB) => NTFunc nonTermA nonTermB -> NTFunc nonTermB nonTermA
invert_nt nt = let step1 = concat . map expand_invert $ nt
               in  foldr (group_pair) [] step1
          where expand_invert (a,l) = map (\x -> (x,[a])) l

-- Auxiliary function
group_pair :: (Eq a) => (a,[b]) -> [(a,[b])] -> [(a,[b])]
group_pair (a,b) [] = [(a,b)]
group_pair ab@(a,b) (h@(a0,b0):t)
           | a == a0   = (a0,b ++ b0):t
           | otherwise = h:(group_pair ab t)

-- Intermediate data type to deal with rules inversion
data TransfAux domain_type target_types canBes domain_pattern target_pattern
			= TransfAux { -- The domain type of this rule
                          domain_type  :: domain_type
						-- All the types the rules of this type
						-- must go to, according to the nt relation
                        , target_types :: target_types
						-- All the canBes that can be applied to
						-- the rules with this domain type
                        , canbes       :: canBes
						-- the domain pattern of this rule
                        , domain_patt  :: domain_pattern
						-- all the patterns this rule can map to.
						-- When the set of the types of these patterns
						-- is equal or greater than target_types, we have
						-- a complete rule
                        , target_patt  :: target_pattern
                        } deriving (Show, Eq)

-- We are assuming there is never repetition among rules, i.e.,
-- two rules (Patt A, Patt B) and (Patt A, Patt C), after the initial inversion, never exist.
-- Rather, there will be: (Patt A, [Patt B, Patt C])
transf2TransfAux :: (Eq nonTermB, Ord nonTermA) => NTFunc nonTermB nonTermA -> [Rule termA nonTermA prodsA termB nonTermB prodsB]
     -> [TransfAux
           nonTermB
           [nonTermA]
           [CanBe termB nonTermB prodsB termA nonTermA prodsA]
           (Pattern termB nonTermB prodsB)
           [Pattern termA nonTermA prodsA]
		]
transf2TransfAux nt rules
    = let first_inversion  = map (\(a,b) -> (b,a)) rules
          canBes           = filter (\x -> case x of
                                          (P_Arg _,_) -> True
                                          otherwise   -> False
                                    ) $ first_inversion
          cleaned_transf   = filter (\x -> case x of
                                          (Pattern _ _,_) -> True
                                          otherwise       -> False
                                    ) $ first_inversion
          -- SHOULD insert step here to check for repeition of rules
		  -- in cleaned_transf (two rules with the same pattern on the lhs)
          create_TransfAux = map (\x -> add_one x nt canBes) cleaned_transf
                           where add_one rule nt canBes = let domain_type     = getPatternType . fst $ rule
		   			                                    -- When calculating the target_types, there will
		   												                  -- be only one, so we can abuse the head function
                                                              target_types    = sort . snd . head . filter ((==domain_type) . fst) $ nt
                                                              canBes_filtered = filter ((==domain_type) . getPatternType . fst) $ canBes
                                                              domain_patt     = fst rule
                                                              target_patt     = [snd rule]
                                                          in  TransfAux domain_type target_types canBes_filtered domain_patt target_patt
      in  create_TransfAux

-- This function can result in an infinite computation or can fail if there are no
-- rules or canBes to force the nt
expandTransfAux :: Ord nonTermB => [TransfAux
                                    nonTermA
                                    [nonTermB]
                                    [CanBe termA nonTermA prodsA termB nonTermB prodsB]
                                    (Pattern termA nonTermA prodsA)
                                    [Pattern termB nonTermB prodsB] ]
                              -> [TransfAux
                                    nonTermA
                                    [nonTermB]
                                    [CanBe termA nonTermA prodsA termB nonTermB prodsB]
                                    (Pattern termA nonTermA prodsA)
                                    [Pattern termB nonTermB prodsB]]
expandTransfAux = map (\x -> aux (map (getPatternType) $ target_patt x) x)
        where aux targettypesTransfAux transfAux 
        	= if (sort targettypesTransfAux) == target_types transfAux
              then transfAux
        	  else let -- getting rid of the canBes that don't matter (whose toType is already present in targettypesTransfAux or
        			   -- whose toType is alreasy present in a canBe). We can have way more canBes/Patterns than those enforced
                       -- in the nt
        	           types_of_target_patts = map (getPatternType) (target_patt transfAux)
        	           canbes_filtered       = filter ((\x -> elem x (targettypesTransfAux++types_of_target_patts)) . canBefromType) (canbes transfAux)
                       -- getting rid of patterns that don't matter   (whose type is already present in targettypesTransfAux or that do not have a canBe
        			   -- relation that can promoted them to other type)
        	           types_of_canBes       =  map (canBefromType) (canbes transfAux)
        	           target_patt_filtered  = filter ((\x -> elem x (targettypesTransfAux++types_of_canBes)) . getPatternType)             (target_patt transfAux)
        		       -- now, we pick a pattern, find all the canBe relations that accept its type, and promote it
        	           pattern_to_promote    = head target_patt_filtered
        	           canBes_to_promote     = filter ((==(getPatternType pattern_to_promote)) . canBefromType) canbes_filtered
        			   -- Do we need an auxiliary step, to make sure there aren't two canBes that promote to the same type ?????
        			   -- Apparently not, but lets keep this note here, for when everything goes terribly wrong.
        	           new_patterns          = map ((\(Pattern p rhs) -> Pattern p (aux rhs pattern_to_promote)) . snd) canBes_to_promote
        	                                 where aux []                   patt = [patt]
        	                                       aux (P_Arg   p     : xs) patt = patt : xs
        	                                       aux (P_Term  p     : xs) patt = (P_Term p)    : aux xs patt
        	                                       aux (P_Lexemme p   : xs) patt = (P_Lexemme p) : aux xs patt
        	                                       aux (Pattern p rhs : xs) patt = (Pattern p (aux rhs patt)) : aux xs patt
        	           newTransfAux          = transfAux {target_patt = new_patterns++(target_patt transfAux)}
        	           newtargettypesTransfAux = (map (getPatternType) new_patterns) ++ targettypesTransfAux
        	       in  aux newtargettypesTransfAux newTransfAux

transfAux2Transf :: NTFunc nonTermA nonTermB
                  -> [TransfAux nonTermA
                               [nonTermB]
                               [CanBe termA nonTermA prodsA termB nonTermB prodsB]
                               (Pattern termA nonTermA prodsA)
                               [Pattern termB nonTermB prodsB]]
		          -> Transf termA nonTermA prodsA termB nonTermB prodsB
transfAux2Transf nt transfAux = Transf nt (concat . map (aux) $ transfAux)
                 where aux transfAux = let patt1  = domain_patt transfAux
                                           patts2 = target_patt transfAux
                                       in  map (\x -> (patt1,x)) patts2

invert :: (Eq nonTermA, Ord nonTermB) =>
     Transf termB nonTermB prodsB termA nonTermA prodsA
     -> Transf termA nonTermA prodsA termB nonTermB prodsB
invert (Transf nt rules) = let inverted_nt    = invert_nt nt
                               inverted_rules =  expandTransfAux (transf2TransfAux inverted_nt rules)
                           in  transfAux2Transf inverted_nt inverted_rules


