module Main where

import Bidi
import Data.List
import Data.Maybe
import Data.Data

import Terminals2Silver
import Source2Abstract
import Source2Concrete
import View2Abstract

r :: IO()
r = do putStrLn "Generating Silver Files..."
       writeFile "Gram/Concrete_Abstract.sv" (source2Abstract concrete_grammar)
       writeFile "Gram/Concrete_Concrete.sv" (source2Concrete concrete_grammar)
       writeFile "Gram/Abstract_Abstract.sv" (view2Abstract   abstract_grammar)
       writeFile "Gram/Terminals.sv"          (initialize)
       putStrLn ("Silver Files generated in './Gram'.")

-- -- -- Concrete Grammar
data Concrete_Terminals = LeftParen | RightParen | Plus | Minus | Ex 
                        deriving (Show, Eq, Ord)
instance RegExp Concrete_Terminals where
	regExp LeftParen  = "("
	regExp RightParen = ")"
	regExp Plus       = "+"
	regExp Minus      = "-"
	regExp Ex         = "*"

data Concrete_Nonterminals = E | T | F
                           deriving (Show, Eq, Ord)
data Concrete_Prod_Names = Add | Sub | Et | Mul | Tf | Neg | Nest | Value 
                         deriving (Show, Eq, Ord)

concrete_grammar :: CFG Concrete_Terminals Concrete_Nonterminals Concrete_Prod_Names
concrete_grammar = CFG [E, T, F] [Int_Lit] [LeftParen, RightParen, Plus, Minus, Ex] [Add, Sub, Et, Mul, Tf, Neg, Nest, Value]
                       [ Prod Add   E [NonTerm E,      Term Plus,  NonTerm T      ]
                       , Prod Sub   E [NonTerm E,      Term Minus, NonTerm T      ]
                       , Prod Et    E [NonTerm T                                  ]
                       , Prod Mul   T [NonTerm T,      Term Ex,    NonTerm F      ]
                       , Prod Tf    T [NonTerm F                                  ]
                       , Prod Neg   F [Term Minus,     NonTerm F                  ]
                       , Prod Nest  F [Term LeftParen, NonTerm E, Term RightParen ]
                       , Prod Value F [Prim Int_Lit                               ]
                       ]

-- -- -- Abstract Grammar
data Abstract_Terminals = Abstract_Nonterminals
                          deriving (Show, Ord, Eq)
data Abstract_Nonterminals = A
                           deriving (Show, Eq, Ord)
instance RegExp Abstract_Terminals where
	regExp _ = ""

data Abstract_Prod_Names = ADDa | SUBa | MULa | VALUEa | ZERO
                         deriving (Show, Eq, Ord)

abstract_grammar :: CFG Abstract_Terminals Abstract_Nonterminals Abstract_Prod_Names
abstract_grammar = CFG [A] [Int_Lit] [] [ADDa, SUBa, MULa, VALUEa]
                       [ Prod ADDa   A [NonTerm A, NonTerm A]
                       , Prod SUBa   A [NonTerm A, NonTerm A]
                       , Prod MULa   A [NonTerm A, NonTerm A]
                       , Prod VALUEa A [Prim Int_Lit        ]
                       ]

-- -- -- Transformation from Concrete to Abstract (T3 Transformation type)
ntCA :: NTFunc Concrete_Nonterminals Abstract_Nonterminals
ntCA = [(E,[A]),(T,[A]),(F,[A])]

rulesCA :: [Rule Concrete_Terminals Concrete_Nonterminals Concrete_Prod_Names 
                 Abstract_Terminals Abstract_Nonterminals Abstract_Prod_Names]
rulesCA = [ (Pattern (Add,E)   [P_Arg ('a',E), P_Term Plus,  P_Arg ('b',T)]         , Pattern (ADDa,A)   [P_Arg ('a',A), P_Arg ('b',A)])
          , (Pattern (Sub,E)   [P_Arg ('a',E), P_Term Minus, P_Arg ('b',T)]         , Pattern (SUBa,A)   [P_Arg ('a',A), P_Arg ('b',A)])
		  , (Pattern (Et,E)    [P_Arg ('a',T)]                                      , P_Arg   ('a',A))
		  , (Pattern (Mul,T)   [P_Arg ('a',T), P_Term Ex,    P_Arg ('b',F)]         , Pattern (MULa,A)   [P_Arg ('a',A), P_Arg ('b',A)])
		  , (Pattern (Tf,T)    [P_Arg ('a',F)]                                      , P_Arg   ('a',A))
		  , (Pattern (Neg,F)   [P_Term Minus, P_Arg ('a',F)]                        , Pattern (SUBa,A)   [Pattern (VALUEa,A) [P_Lexemme ('i',"0", Int_Lit)], P_Arg ('a',A)])
		  , (Pattern (Nest,F)  [P_Term LeftParen, P_Arg ('a',E), P_Term RightParen] , P_Arg   ('a',A))
          , (Pattern (Value,F) [P_Prim ('a',Int_Lit)]                               , Pattern (VALUEa,A) [P_Prim ('a',Int_Lit)])
          ]

getCA :: Transf Concrete_Terminals Concrete_Nonterminals Concrete_Prod_Names
                Abstract_Terminals Abstract_Nonterminals Abstract_Prod_Names
getCA =  Transf ntCA rulesCA

-- -- -- This is the transformation from the Abstract to the Concrete.
--       It is automatically generated by the inversion engine
putAC :: Transf Abstract_Terminals Abstract_Nonterminals Abstract_Prod_Names
                Concrete_Terminals Concrete_Nonterminals Concrete_Prod_Names
putAC = Transf {
  transfNT = [(A,[E,T,F])]
  , transfRules = 
     [(Pattern (ADDa,A) [P_Arg ('a',A),P_Arg ('b',A)],Pattern (Tf,T) [Pattern (Nest,F) [P_Term LeftParen,Pattern (Add,E) [P_Arg ('a',E),P_Term Plus,P_Arg ('b',T)],P_Term RightParen]])
    ,(Pattern (ADDa,A)   [P_Arg ('a',A),P_Arg ('b',A)],Pattern (Nest,F) [P_Term LeftParen,Pattern (Add,E) [P_Arg ('a',E),P_Term Plus,P_Arg ('b',T)],P_Term RightParen])
	,(Pattern (ADDa,A)   [P_Arg ('a',A),P_Arg ('b',A)],Pattern (Add,E) [P_Arg ('a',E),P_Term Plus,P_Arg ('b',T)])
	,(Pattern (SUBa,A)   [P_Arg ('a',A),P_Arg ('b',A)],Pattern (Tf,T) [Pattern (Nest,F) [P_Term LeftParen,Pattern (Sub,E) [P_Arg ('a',E),P_Term Minus,P_Arg ('b',T)],P_Term RightParen]])
	,(Pattern (SUBa,A)   [P_Arg ('a',A),P_Arg ('b',A)],Pattern (Nest,F) [P_Term LeftParen,Pattern (Sub,E) [P_Arg ('a',E),P_Term Minus,P_Arg ('b',T)],P_Term RightParen])
	,(Pattern (SUBa,A)   [P_Arg ('a',A),P_Arg ('b',A)],Pattern (Sub,E) [P_Arg ('a',E),P_Term Minus,P_Arg ('b',T)])
	,(Pattern (MULa,A)   [P_Arg ('a',A),P_Arg ('b',A)],Pattern (Nest,F) [P_Term LeftParen,Pattern (Et,E) [Pattern (Mul,T) [P_Arg ('a',T),P_Term Ex,P_Arg ('b',F)]],P_Term RightParen])
	,(Pattern (MULa,A)   [P_Arg ('a',A),P_Arg ('b',A)],Pattern (Et,E) [Pattern (Mul,T) [P_Arg ('a',T),P_Term Ex,P_Arg ('b',F)]])
	,(Pattern (MULa,A)   [P_Arg ('a',A),P_Arg ('b',A)],Pattern (Mul,T) [P_Arg ('a',T),P_Term Ex,P_Arg ('b',F)])
	,(Pattern (SUBa,A)   [Pattern (VALUEa,A)  [P_Lexemme ('i',"0",Int_Lit)],P_Arg ('a',A)], Pattern (Et,E) [Pattern (Tf,T) [Pattern (Neg,F) [P_Term Minus,P_Arg ('a',F)]]])
	,(Pattern (SUBa,A)   [Pattern (VALUEa,A)  [P_Lexemme ('i',"0",Int_Lit)],P_Arg ('a',A)], Pattern (Tf,T) [Pattern (Neg,F) [P_Term Minus,P_Arg ('a',F)]])
	,(Pattern (SUBa,A)   [Pattern (VALUEa,A)  [P_Lexemme ('i',"0",Int_Lit)],P_Arg ('a',A)], Pattern (Neg,F) [P_Term Minus,P_Arg ('a',F)])
	,(Pattern (VALUEa,A) [P_Prim ('a',Int_Lit)],Pattern (Et,E) [Pattern (Tf,T) [Pattern (Value,F) [P_Prim ('a',Int_Lit)]]])
	,(Pattern (VALUEa,A) [P_Prim ('a',Int_Lit)],Pattern (Tf,T) [Pattern (Value,F) [P_Prim ('a',Int_Lit)]])
	,(Pattern (VALUEa,A) [P_Prim ('a',Int_Lit)],Pattern (Value,F) [P_Prim ('a',Int_Lit)])
    ]
}

