{-# LANGUAGE TypeSynonymInstances, FlexibleInstances #-}

module GetTransf2Silver where

import Bidi
import Data.List


-- -- ****** This module is EXACLY the same as PutTransf2Silver.hs, but instead of creating a Silver AG with
-- -- an attribute called putTransf, it creates one with a module names getTransf.
-- -- The inherent mechanism is exactly the same and should continue that way.
-- -- Because we will overlapping instances, we had to rename the class toSilver to toSilverGet


-- This Module looks more complicated than it seems, so for each instance of toSilver
-- I will explain in detail what each instance does.

-- First, we use the function transf2SilverTransf to convert our transformation datatype
-- into something wasy to manage. Please refer to the function to see what it does
-- (basically, we sort everything into lists of lists, sorted first by their production,
-- and then by their length).
instance (Ord termA, Ord nonTermA, Ord prod_nameA, Ord termB, Ord nonTermB, Ord prod_nameB,
          Show nonTermB, Show nonTermA, Show prod_nameA, Show prod_nameB, Show termB,
          RegExp termB, Show termA, RegExp termA) =>
         CreateGet (Transf termA nonTermA prod_nameA termB nonTermB prod_nameB) where
  toSilverGet = toSilverGet . transf2SilverTransf

-- This instance initializes the attributes declaration
instance (Show nonTermB, Show nonTermA, Show prod_nameA, Show prod_nameB, Show termB, RegExp termB,
          Show termA, RegExp termA) =>
         CreateGet (IntermediateTransf termA nonTermA prod_nameA termB nonTermB prod_nameB) where
  toSilverGet transf =
          -- by default all grammars are called Gram
      let header = "grammar Gram;\n"
	      -- attrs is the attributes declarations
          attrs  = toSilverGet (interTransfNT transf)
		  -- prods are the attributes specifications
          prods  = concat . map toSilverGet $ (orderedRules transf)
      in  unlines [header,attrs, prods]

-- This instance handles the attributes declarations
instance (Show nonTermB, Show nonTermA) => CreateGet (NTFunc nonTermA nonTermB) where
  toSilverGet nt = let attributes = unlines . nub . map (\(a,b) -> auxAttrs b) $ nt
                       occurences = unlines . nub . map (\(a,l) -> auxOcc a l) $ nt
                   in  unlines [attributes, occurences]
         where auxAttrs = concat . map (\x -> "synthesized attribute get_"++show x++" :: "++show x++";")
               auxOcc a = concat . map (\x -> "attribute get_"++show x++" occurs on "++show a++";")

-- This instance initializes the productions where the attributes will be specified
-- (the input lists have rules whose lhs productions are exactly the same)
instance (Show prod_nameA, Show nonTermB, Show prod_nameB, Show termB, RegExp termB,
          Show nonTermA, Show termA, RegExp termA) =>
         CreateGet [[Rule termA nonTermA prod_nameA termB nonTermB prod_nameB]] where
  toSilverGet l =
      let a1       = "aspect production p_" ++ (show . getPatternProd . fst . head . head $ l)
-- **** VERY IMPORTANTE: for this function to work, it is assumed there ALWAYS is a default
-- rule in the transformation, i.e., a rule that maps directly to the production's
-- children, with no need to pattern match. This rule is used to define the abstract
-- production in Silver, such as: this::A ::= a::A b::A
          a2       = (\(Pattern (prod,nt) l) -> "this::"++show nt++" ::= "++(intercalate " " $ map aux l)) . fst . last . head $ l
          listVars = (\(Pattern (prod,nt) l) -> getListVars l) . fst . last . head $ l
          a3       = concat . map toSilverGet $ map (\x -> (x,listVars)) l
      in  unlines [a1,a2,"{",a3,"}"]
      where aux (P_Term p)          = show p
            aux (P_Prim (n,t))      = [n]++"::"++show t
            aux (P_Arg  (n,t))      = [n]++"::"++show t
			-- For why this gives an error, check the comment below
            aux (Pattern (n,t) rhs) = error "*** For this error, refer to Transf2Silver.hs, line 59 ***"
	  

-- This instance initializes the attributes per non-terminal
-- and calls instances to create each local attribute
-- The "error" attribute is also created here
instance (Show prod_nameA, Show nonTermB, Show prod_nameB, Show termB, RegExp termB,
          Show termA, Show nonTermA, RegExp termA) =>
         CreateGet ([Rule termA nonTermA prod_nameA termB nonTermB prod_nameB], [Char]) where
  toSilverGet (l,listVars)
     =  let lengthR = length l
            nonTerm = show . getPatternType . snd . head $ l
            a1      = "  this.get_"++nonTerm++" = get_"++nonTerm++"_"++show lengthR++";"
            -- The zip is to know which local attribute I am creating when 
            -- implementing 'map toSilver'
            a2      = concat . map toSilverGet $ addList (zip (reverse [1..lengthR]) l) listVars
            a3      = "  local attribute get_"++nonTerm++"_0 :: "++nonTerm++" = error (\"This transformation is not total for non terminal "++nonTerm++"\");"
         in  unlines [a1,a2,a3,"\n"]

-- This instance creates a list of 
instance (Show prod_nameB, Show termB, RegExp termB, Show nonTermB,
          Show termA, Show nonTermA, Show prod_nameA, RegExp termA) =>
         CreateGet (Int, Rule termA nonTermA prod_nameA termB nonTermB prod_nameB, [Char]) where
  toSilverGet (n,r, listVars)
     = let nonTerm       = show . getPatternType . snd $ r
           listlexes     = getListLexemes (snd r)
           a1            = "  local attribute get_"++nonTerm++"_"++show n++" :: "++nonTerm++" = "
           -- matchingPatt creates the cases to pattern match subtree.
           -- The arguments are: pattern to be matched (fst r), the pattern
           -- it is transformed into, in case it matches, in string form,
           -- and the local attr the case goes to in case the matching 
           -- fails. The last int is only for identation
           matchingPatt  = pattern2Cases (fst r) resultingPatt ("get_"++nonTerm++"_"++show (n-1)) 27 listVars
           resultingPatt = printTargetPattern (snd r)
       in  unlines [a1++matchingPatt]

-- This datatype has exactly the same information as the Transf datatype, but
-- list of rules was preprocessed by transf2SilverTransf
data IntermediateTransf termA nonTermA prod_nameA termB nonTermB prod_nameB
  = IntermediateTransf { interTransfNT     :: NTFunc nonTermA nonTermB,
                         orderedRules      :: [[[ Rule termA nonTermA prod_nameA termB nonTermB prod_nameB ]]] }
                       deriving (Show, Eq, Ord)

-- Function that translates Transf to IntermediateTransf
-- In the end of this function we will end with:
-- [ list of rules [list of rules by their production [list of rules by their target nonterminal, sorted by length]] ]
transf2SilverTransf :: (Ord termA, Ord nonTermA, Ord prod_nameA, Ord termB, Ord nonTermB, Ord prod_nameB)
                    => (Transf termA nonTermA prod_nameA termB nonTermB prod_nameB)
                    -> (IntermediateTransf termA nonTermA prod_nameA termB nonTermB prod_nameB)
transf2SilverTransf t = let nt    = transfNT t
                            -- In this first step, we group the rules by their
							-- lhs pattern prod. For groupBy to work, the list
							-- must be sorted first.
                            rules = groupBy lhsGroup . sortBy lhsSort . transfRules $ t
                            -- Same thing as he previous step, only this time we
                            -- group rules by their rhs pattern type (remember,
                            -- a pattern type is the nonterminal on topmost)
                            -- The last step is sorting rules by the size of their
							-- lhs pattern
                            rulesByNT = map (map (sortBy lengthSort)) . map (groupBy rhsGroup) . map (sortBy rhsSort) $ rules
                        in  IntermediateTransf {interTransfNT = nt, orderedRules = rulesByNT}
  where lhsSort a b | getPatternProd (fst a) > getPatternProd (fst b)  = GT
                    | getPatternProd (fst a) < getPatternProd (fst b)  = LT
                    | getPatternProd (fst a) == getPatternProd (fst b) = EQ
        rhsSort a b | getPatternType (snd a) > getPatternType (snd b)  = GT
                    | getPatternType (snd a) < getPatternType (snd b)  = LT
                    | getPatternType (snd a) == getPatternType (snd b) = EQ
        lengthSort a b | getPatternDepth (fst a) < getPatternDepth (fst b) = GT
                       | getPatternDepth (fst a) > getPatternDepth (fst b) = LT
                       | otherwise                                         = EQ
        lhsGroup a b = getPatternProd (fst a) == getPatternProd (fst b)
        rhsGroup a b = getPatternType (snd a) == getPatternType (snd b)


-- In the instance: CreatePut (Int, Rule termA nonTermA prod_nameA termB nonTermB prod_nameB)
-- the arguments of this function are explained.
-- listVars is generated by getListVars
pattern2Cases :: (Show prodName, Show nonTerm, Show term, RegExp term) =>
                 Pattern term nonTerm prodName -> String -> String -> Int -> [Char] -> String
pattern2Cases patt@(Pattern (p,t) rhs) g b ident listVars =
  let childs       = concat (map getKids rhs)
      numberChilds = length childs
      a1           = "case "++(intercalate ", " $ map (\x -> [x]) $ take numberChilds listVars)++" of"
      a2           = spaces ident++"   | "++intercalate ", " (map printSourcePattern childs)++" -> "++drawIfs (getLexemmes patt) g b
	  -- We test if all the childs are P_Args here because, it that is true, Silver will
	  -- give an "Pattern has overlapping cases" error, so if there are only P_Args, we will
	  -- not need a default case with underscores. Nevertheless, we keep the loal attribute
	  -- that generates an error because it might be needed for other cases.
      a3           = if (areAllChildsP_Args rhs) then ""
                     else "\n"++spaces ident++"   | "++ (intercalate ", " (under numberChilds)) ++ " -> " ++ b
      a4           = spaces ident++"end;"
  in  if (numberChilds == 0) then g++";" else unlines [a1,a2++a3,a4]
  where getKids     (P_Term  p   )        = []
        getKids     (P_Prim (n,t))        = []
        getKids     (P_Arg  (n,t))        = [P_Arg  (n,t)]
        getKids     (Pattern (p,t) rhs)   = [Pattern (p,t) rhs]
        getKids     (P_Lexemme (n,p,t)  ) = []
        getLexemmes (P_Term  p   )        = []
        getLexemmes (P_Prim (n,t))        = []
        getLexemmes (P_Arg  (n,t))        = []
        getLexemmes (Pattern (p,t) rhs)   = concat (map getLexemmes rhs)
        getLexemmes (P_Lexemme (n,p,t)  ) = [(n,p)]
        drawIfs [] g _ = g
        drawIfs l  g b = let lexemesTests = concat $ map (\(a,b) -> [a]++".lexeme == \""++b++"\"") l
                         in  "if ("++lexemesTests++") then "++g++" else "++b
        areAllChildsP_Args [] = True
        areAllChildsP_Args (P_Term  p         :xs) = True  && areAllChildsP_Args xs
        areAllChildsP_Args (P_Prim (n,t)      :xs) = True  && areAllChildsP_Args xs
        areAllChildsP_Args (P_Arg  (n,t)      :xs) = True  && areAllChildsP_Args xs
        areAllChildsP_Args (Pattern (p,t) rhs :xs) = False && areAllChildsP_Args xs
        areAllChildsP_Args (P_Lexemme (n,p,t) :xs) = False && areAllChildsP_Args xs



-- This function gets a Pattern and creates a symbol table composed
-- by the P_Lexeme's, and the value the terminal must have
-- We are talking about the 'Lexeme' terminals here, which we have
-- to test with if then else (cases do not work).
getListLexemes :: Pattern term nonTerm prod_name -> [(Char,String)]
getListLexemes (P_Term    p        ) = []
getListLexemes (P_Prim    (n,t)    ) = []
getListLexemes (P_Arg     (n,t)    ) = []
getListLexemes (Pattern   (p,t) rhs) = concat . map getListLexemes $ rhs
getListLexemes (P_Lexemme (n,p,t)  ) = [(n,p)]

-- This function takes the target pattern and transformas it into a the resulting
-- subtree, and creates the recursive calls to this tree non-terminals
printTargetPattern :: (RegExp term, Show term, Show nonTerm, Show prodName) => (Bidi.Pattern term nonTerm prodName) -> String
printTargetPattern (P_Term  p   )        = "terminal ("++show p++", \""++regExp p++"\")"
printTargetPattern (P_Prim (n,t))        = [n]
printTargetPattern (P_Arg  (n,t))        = [n]++".get_"++(show t)
printTargetPattern (Pattern (p,t) rhs)   = "p_"++(show p)++"( "++(intercalate ", " $ map printTargetPattern rhs)++" )"
printTargetPattern (P_Lexemme (n,p,t)  ) = "terminal ("++show t++", \""++p++"\")"

-- This function takes the target pattern and transformas it into a the resulting
-- subtree, and creates the recursive calls to this tree non-terminals
printSourcePattern :: (RegExp term, Show term, Show nonTerm, Show prodName) => (Bidi.Pattern term nonTerm prodName) -> String
printSourcePattern (P_Arg  (n,t))        = "_"
printSourcePattern (Pattern (p,t) rhs)   = "p_"++(show p)++"( "++(intercalate ", " $ map printSourcePattern rhs)++" )"
printSourcePattern (P_Lexemme (n,p,t)  ) = [n]
printSourcePattern (P_Term  p   )        = "_"
--printSourcePattern (P_Prim (n,t))        = [n]

-- This function gets a pattern and extracts a list with all its P_Arg variables, in the
-- right order
getListVars :: [Pattern t t1 t2] -> [Char]
getListVars [] = ""
getListVars (P_Term p    :xs)      = getListVars xs
getListVars (P_Prim (n,t):xs)      = n : getListVars xs
getListVars (P_Arg  (n,t):xs)      = n : getListVars xs
-- For why this gives an error, check the comment below
--getListVars (Pattern (n,t) rhs) = error "*** For this error, refer to Transf2Silver.hs, line 59 ***"

-- Aux funcs
under n = take n aux
   where aux = "_":aux
spaces n = take n aux
   where aux = ' ':aux
addList :: [(a,b)] -> [c] -> [(a,b,[c])]
addList [] _ = []
addList ((a,b):xs) l = (a,b,l) : addList xs l


















