{-# LANGUAGE TypeSynonymInstances, FlexibleInstances #-}

module Types where

import Data.List

-- -- -- Transformation datatypes
data Pattern termA nonTermA prodNameA termB nonTermB prodNameB 
  = P_Term    termA
  | P_Prim    Char Prim
  | P_Arg     Char nonTermA
  | P_Lexemme Char String Prim -- For example: (P_Lexeme 'i' "2" Int_Lit)
  | Pattern   prodNameA nonTermA [Pattern termA nonTermA prodNameA termB nonTermB prodNameB]
  | P_Get     nonTermA (Pattern termB nonTermB prodNameB termA nonTermA prodNameA)
  | P_Put     nonTermA (Pattern termB nonTermB prodNameB termA nonTermA prodNameA)
  deriving (Eq, Ord)

type Rule termA nonTermA prodNameA termB nonTermB prodNameB = ( Pattern termA nonTermA prodNameA termB nonTermB prodNameB
                                                              , Pattern termA nonTermA prodNameA termB nonTermB prodNameB)

data Transf termA nonTermA prodNameA termB nonTermB prodNameB = Transf [Rule termA nonTermA prodNameA termB nonTermB prodNameB]

get :: nonTermA -> nonTermB -> Pattern termA nonTermA prodNameA termB nonTermB prodNameB -> Pattern termB nonTermB prodNameB termA nonTermA prodNameA
get ntA ntB prodA = P_Get ntB prodA

-- -- -- Context-free Grammar datatype
data CFG term nonTerm prodName = CFG { cfgNonTerms  :: [nonTerm]
                                      , cfgPrims     :: [Prim]
                                      , cfgTerms     :: [term]
                                      , cfgProdNames :: [prodName]
                                      , cfgProds     :: [Prod prodName nonTerm term]
                                      } deriving (Show, Eq, Ord)

data Prod prodName nonTerm term = Prod { prodName     :: prodName
                                       , prodNonTerm  :: nonTerm
                                       , prodRhs      :: [RHS nonTerm term]
                                       } deriving (Show, Eq, Ord)

data RHS nonTerm term = Term term | NonTerm nonTerm | Prim Prim
                      deriving (Show, Eq, Ord)

-- -- -- Auxiliary stuff
data Prim = Int_Lit | String_Lit | Char_Lit
           deriving (Eq, Show, Ord)

class RegExp a where
  regExp :: a -> String

instance RegExp Prim where
  regExp String_Lit = "/[a-zA-Z][a-zA-Z0-9_]*/"
  regExp Char_Lit   = "/[a-zA-Z][a-zA-Z0-9_]*/"

spaces n = take n aux
  where aux = ' ':aux

-- -- -- Instantiations of Show for Transf and CFG
instance (RegExp termA, RegExp termB, Show nonTermA, Show nonTermB, Show prodNameA, Show prodNameB)
         => Show (Pattern termA nonTermA prodNameA termB nonTermB prodNameB) where
  show (P_Term termA)                        = "'" ++ regExp termA ++ "'"
  show (P_Prim char prim)                    = (show prim) ++ " '" ++ [char] ++ "'"
  show (P_Arg char nonTermA)                 = (show nonTermA) ++ " '" ++ [char] ++ "'"
  show (P_Lexemme char value prim)           = "lexeme " ++ show prim ++ " '" ++ [char] ++ "' " ++ value
  show (Pattern prodNameA nonTermA patterns) = show nonTermA++" ::= "++show prodNameA ++ "( " ++ intercalate ", " (map show patterns) ++ " )"
  show (P_Get nonTermA pattern)              = "( " ++ show pattern ++ " ).get"++show nonTermA
  show (P_Put nonTermA pattern)              = "( " ++ show pattern ++ " ).put"++show nonTermA

instance (RegExp termA, RegExp termB, Show nonTermA, Show nonTermB, Show prodNameA, Show prodNameB)
         => Show (Transf termA nonTermA prodNameA termB nonTermB prodNameB) where
  show (Transf l) = let max_lhs = maximum $ map (length . show . fst) l
                    in  unlines $ map (aux max_lhs) l
                    where aux max_lhs (p1,p2) = let show_p1 = show p1
                                                    show_p2 = show p2
                                                    indent  = spaces (max_lhs - (length show_p1))
                                                in  show_p1 ++ indent ++ " -> " ++ show_p2

-- -- -- Auxiliary functions for Patterns
getPatternVar :: Pattern termA nonTermA prodNameA termB nonTermB prodNameB -> Char
getPatternVar (Pattern  _  _ _   ) = error "Pattern have no variables!"
getPatternVar (P_Arg    v _      ) = v
getPatternVar (P_Prim   v _      ) = v
getPatternVar (P_Term  _         ) = error "P_Term have no variables!"
getPatternVar (P_Lexemme  v _ _  ) = v
getPatternVar (P_Get   _ _       ) = error "P_Get have no variables!"
getPatternVar (P_Put   _ _       ) = error "P_Put have no variables!"

getPatternType :: Pattern termA nonTermA prodNameA termB nonTermB prodNameB -> nonTermA
getPatternType (Pattern  _ t _    ) = t
getPatternType (P_Arg    _ t      ) = t
-- Prims DO have types. They're not non-terminals though, so we just generate an error.
getPatternType (P_Prim   _ _      ) = error "Someone, somewhere is calling the function getPatternType on a P_Prim!"
getPatternType (P_Term  _         ) = error "Someone, somewhere is calling the function getPatternType on a P_Term!"
getPatternType (P_Lexemme  _ _ _  ) = error "Someone, somewhere is calling the function getPatternType on a P_Lexemme!"
getPatternType (P_Get   t _       ) = t
getPatternType (P_Put   t _       ) = t

getPatternProd :: Pattern termA nonTermA prodNameA termB nonTermB prodNameB -> prodNameA
getPatternProd (Pattern  p  _ _   ) = p
getPatternProd (P_Arg    _ _      ) = error "Someone, somewhere is calling the function getPatternProd on a P_Arg!"
getPatternProd (P_Prim   _ _      ) = error "Someone, somewhere is calling the function getPatternProd on a P_Prim!"
getPatternProd (P_Term   _        ) = error "Someone, somewhere is calling the function getPatternProd on a P_Term!"
getPatternProd (P_Lexemme  _ _ _  ) = error "Someone, somewhere is calling the function getPatternProd on a P_Lexemme!"
getPatternProd (P_Get   _ p       ) = error "Someone, somewhere is calling the function getPatternProd on a P_Get!"
getPatternProd (P_Put   _ p       ) = error "Someone, somewhere is calling the function getPatternProd on a P_Put!"

getPatternRHS :: Pattern termA nonTermA prodNameA termB nonTermB prodNameB -> [Pattern termA nonTermA prodNameA termB nonTermB prodNameB]
getPatternRHS (Pattern _ _ rhs ) = rhs
getPatternRHS (P_Arg   _ _     ) = error "Someone, somewhere is calling the function getPatternRHS on a P_Arg!"
getPatternRHS (P_Prim  _ _     ) = error "Someone, somewhere is calling the function getPatternRHS on a P_Prim!"
getPatternRHS (P_Term  _       ) = error "Someone, somewhere is calling the function getPatternRHS on a P_Term!"
getPatternRHS (P_Lexemme _ _ _ ) = error "Someone, somewhere is calling the function getPatternRHS on a P_Lexemme!"
getPatternRHS (P_Get  _ p      ) = error "Someone, somewhere is calling the function getPatternRHS on a P_Get!"
getPatternRHS (P_Put  _ p      ) = error "Someone, somewhere is calling the function getPatternRHS on a P_Put!"

getPatternDepth :: Pattern termA nonTermA prodNameA termB nonTermB prodNameB -> Int
getPatternDepth (Pattern _ _ rhs ) = 1 + if length rhs == 0 then 0 else (maximum (map (getPatternDepth) rhs))
getPatternDepth (P_Arg   _ _     ) = 0
getPatternDepth (P_Prim  _ _     ) = 0
getPatternDepth (P_Term  _       ) = 0
getPatternDepth (P_Lexemme _ _ _ ) = 1
getPatternDepth (P_Get  _ p      ) = getPatternDepth p
getPatternDepth (P_Put  _ p      ) = getPatternDepth p








